/* favorites.js
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * SPDX-License-Identifier: GPL-2.0-or-later
 */

/* exported getExtension */

const Lang = imports.lang;

const Clutter = imports.gi.Clutter;
const Shell = imports.gi.Shell;
const St = imports.gi.St;

const ExtensionUtils = imports.misc.extensionUtils;

const Main = imports.ui.main;
const PanelMenu = imports.ui.panelMenu;
const PopupMenu = imports.ui.popupMenu;

const Me = ExtensionUtils.getCurrentExtension();

const Convenience = Me.imports.convenience;

const _ = Convenience.getTranslations();

const FAVORITES_ALL = 1;
const FAVORITES_ICON = 1;
const FAVORITES_TEXT = 2;

const KEY_FAVORITE_APPS = 'favorite-apps';
const KEY_FAVORITES_SCOPE = 'favorites-scope';
const KEY_SHOW_FAVORITES = 'show-favorites';

const PopupFavoriteMenuItem = new Lang.Class({
    Name: 'PopupFavoriteMenuItem',
    Extends: PopupMenu.PopupImageMenuItem,

    _init(app) {
        this.parent(app.get_name(), app.get_app_info().get_icon());
        this.label.set({ y_expand: true, y_align: Clutter.ActorAlign.CENTER });
        this._app = app;
    },

    activate(event) {
        let modifiers = event ? event.get_state() : 0;
        let isCtrlPressed = (modifiers & Clutter.ModifierType.CONTROL_MASK) !== 0;
        let openNewWindow = this._app.can_open_new_window() && this._app.state === Shell.AppState.RUNNING && (isCtrlPressed);
        if (openNewWindow) {
            this._app.open_new_window(-1);
        } else {
            this._app.activate();
        }
        Main.overview.hide();
        this.parent(event);
    }
});

const FavoritesMenuButton = new Lang.Class({
    Name: 'FavoritesMenuButton',
    Extends: PanelMenu.Button,

    _init() {
        this.parent(1.0, _("Favorites"), false);
        this.menu.actor.add_style_class_name('dschaerf-menu');
        this._hbox = new St.BoxLayout({ style_class: 'panel-status-menu-box' });
        this._icon = new St.Icon({ icon_name: 'emblem-favorite-symbolic', style_class: 'system-status-icon' });
        this._hbox.add_child(this._icon);
        this._label = new St.Label({ text: _("Favorites"), y_expand: true, y_align: Clutter.ActorAlign.CENTER });
        this._hbox.add_child(this._label);
        this._hbox.add_child(PopupMenu.arrowIcon(St.Side.BOTTOM));
        this.actor.add_child(this._hbox);
        this._settings = Convenience.getSettings();
        this._settingsChangedId = [
            this._settings.connect('changed::%s'.format(KEY_FAVORITE_APPS), this._reloadFavorites.bind(this)),
            this._settings.connect('changed::%s'.format(KEY_FAVORITES_SCOPE), this._reloadFavorites.bind(this)),
            this._settings.connect('changed::%s'.format(KEY_SHOW_FAVORITES), this._onChangedShowFavorites.bind(this)),
        ];
        this._onChangedShowFavorites();
        this._appSystem = Shell.AppSystem.get_default();
        this._appSystemChangedId = [
            this._appSystem.connect('app-state-changed', this._reloadFavorites.bind(this)),
            this._appSystem.connect('installed-changed', this._reloadFavorites.bind(this)),
        ]
        this._globalSettingsChangedId = global.settings.connect('changed::%s'.format(KEY_FAVORITE_APPS), this._reloadFavorites.bind(this));
        this._reloadFavorites();
    },

    _onChangedShowFavorites() {
        let showFavorites = this._settings.get_enum(KEY_SHOW_FAVORITES);
        this._icon.visible = showFavorites !== FAVORITES_TEXT;
        this._label.visible = showFavorites !== FAVORITES_ICON;
    },

    _reloadFavorites() {
        this.menu.removeAll();
        let appList = [];
        let favorites = this._settings.get_strv(KEY_FAVORITE_APPS)
        favorites.forEach(desktopFile => {
            let app = this._appSystem.lookup_app(desktopFile);
            if (app) appList.push(app);
        });
        if (this._settings.get_enum(KEY_FAVORITES_SCOPE) === FAVORITES_ALL) {
            let favoriteApps = global.settings.get_strv(KEY_FAVORITE_APPS);
            favoriteApps.forEach(desktopFile => {
                let app = this._appSystem.lookup_app(desktopFile);
                if (app && appList.indexOf(app) < 0) appList.push(app);
            });
        }
        appList.sort((a, b) => a.compare_by_name(b));
        appList.forEach(app => {
            this.menu.addMenuItem(new PopupFavoriteMenuItem(app));
        });
        this.actor.visible = this.menu.numMenuItems > 0;
    },

    destroy() {
        global.settings.disconnect(this._globalSettingsChangedId);
        this._appSystemChangedId.forEach(id => {
            this._appSystem.disconnect(id);
        });
        this._settingsChangedId.forEach(id => {
            this._settings.disconnect(id);
        });
        this.parent();
    }
});

function getExtension() {
    return Main.panel.addToStatusArea('PanelManagerFavoritesMenu', new FavoritesMenuButton(), 1, 'left');
}
