/* places.js
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * SPDX-License-Identifier: GPL-2.0-or-later
 */

/* exported getExtension */

const Lang = imports.lang;

const Clutter = imports.gi.Clutter;
const St = imports.gi.St;

const ExtensionUtils = imports.misc.extensionUtils;

const Main = imports.ui.main;
const PanelMenu = imports.ui.panelMenu;
const PopupMenu = imports.ui.popupMenu;

const Me = ExtensionUtils.getCurrentExtension();

const Convenience = Me.imports.convenience;
const PlacesManager = Me.imports.placesManager;
const UserDialog = Me.imports.userDialog;

const _ = Convenience.getTranslations();

const BOOKMARKS_MAINMENU = 0;

const KEY_BOOKMARKS_POSITION = 'bookmarks-position';
const KEY_SHOW_PLACES = 'show-places';

const PANELMENU_BUTTON_ICON = 1;
const PANELMENU_BUTTON_TEXT = 2;

const PopupPlaceMenuItem = new Lang.Class({
    Name: 'PopupPlaceMenuItem',
    Extends: PopupMenu.PopupImageMenuItem,

    _init(info) {
        this.parent(info.name, info.icon);
        this.label.set({ x_expand: true, y_expand: true, y_align: Clutter.ActorAlign.CENTER });
        if (info.can_eject) {
            this._addButton({
                icon: 'media-eject-symbolic',
                content: 'ejectVolume',
                value: info.name,
                signal: 'eject',
                callback: () => { info.eject(); },
            });
        } else if (info.can_empty) {
            this._addButton({
                icon: 'edit-clear-all-symbolic',
                content: 'clearRecentDocuments',
                value: null,
                signal: 'empty',
                callback: () => { info.empty(); },
            });
        } else if (info.can_remove) {
            this._addButton({
                icon: 'edit-clear-symbolic',
                content: 'removeRecentDocument',
                value: info.name,
                signal: 'remove',
                callback: () => { info.remove(); },
            });
        }
        this._info = info;
    },

    _addButton(params) {
        this.button = new St.Button({ style_class: 'system-menu-action popup-menu-button' });
        this.button.set_child(new St.Icon({ icon_name: params.icon }));
        this.button.connect('clicked', event => {
            let dialog = new UserDialog.ModalUserDialog(params.content, params.value);
            dialog.connect(params.signal, params.callback);
            Main.overview.hide();
            dialog.open();
            this.emit('activate', event);
        });
        this.actor.add_child(this.button);
    },

    activate(event) {
        Main.overview.hide();
        this._info.launch(event.get_time());
        this.parent(event);
    }
});

const PopupPlacesMenuSection = new Lang.Class({
    Name: 'PopupPlacesMenuSection',
    Extends: PopupMenu.PopupMenuSection,

    _init(manager) {
        this.parent();
        this._placesManager = new PlacesManager[manager]();
        this._placesManagerChangedId = this._placesManager.connect('updated', this._reload.bind(this));
        this._reload();
    },

    _addPlaces(menu, kind) {
        let places = this._placesManager.get(kind);
        places.forEach(info => {
            menu.addMenuItem(new PopupPlaceMenuItem(info));
        });
    },

    _reload() {
        this.actor.visible = this.numMenuItems > 0;
    },

    destroy() {
        this._placesManager.disconnect(this._placesManagerChangedId);
        this._placesManager.destroy();
        this.parent();
    }
});

const PopupPlacesBookmarksMenuSection = new Lang.Class({
    Name: 'PopupPlacesBookmarksMenuSection',
    Extends: PopupPlacesMenuSection,

    _init() {
        this._settings = Convenience.getSettings();
        this._settingsChangedId = this._settings.connect('changed::%s'.format(KEY_BOOKMARKS_POSITION), this._reload.bind(this));
        this.parent('PlacesBookmarksManager');
    },

    _reload() {
        let item;
        this.removeAll();
        this._addPlaces(this, 'special');
        if (this._settings.get_enum(KEY_BOOKMARKS_POSITION) === BOOKMARKS_MAINMENU) {
            this._addPlaces(this, 'default');
        } else {
            item = new PopupMenu.PopupSubMenuMenuItem(_("Bookmarks"), true);
            item.icon.icon_name = 'user-bookmarks';
            this._addPlaces(item.menu, 'default');
            if (item.menu.numMenuItems > 0) this.addMenuItem(item);
        }
        item = new PopupMenu.PopupSubMenuMenuItem(_("Bookmarks (custom)"), true);
        item.icon.icon_name = 'user-bookmarks';
        this._addPlaces(item.menu, 'custom');
        if (item.menu.numMenuItems > 0) this.addMenuItem(item);
        this.parent();
    },

    destroy() {
        this._settings.disconnect(this._settingsChangedId);
        this.parent();
    }
});

const PopupPlacesRecentMenuSection = new Lang.Class({
    Name: 'PopupPlacesRecentMenuSection',
    Extends: PopupPlacesMenuSection,

    _init() {
        this.parent('PlacesRecentManager');
    },

    _reload() {
        this.removeAll();
        if (this._placesManager.size > 0) {
            this._addPlaces(this, 'recent');
            let item = new PopupMenu.PopupSubMenuMenuItem(_("Last used documents"), true);
            item.icon.icon_name = 'document-open-recent';
            this._addPlaces(item.menu, 'recent-item');
            this.addMenuItem(item);
        }
        this.actor.visible = this._placesManager.size > 0;
    }
});

const PopupPlacesVolumesMenuSection = new Lang.Class({
    Name: 'PopupPlacesVolumesMenuSection',
    Extends: PopupPlacesMenuSection,

    _init() {
        this.parent('PlacesVolumesManager');
    },

    _reload() {
        let item;
        this.removeAll();
        this._addPlaces(this, 'special');
        item = new PopupMenu.PopupSubMenuMenuItem(_("Network Drives"), true);
        item.icon.icon_name = 'network-server';
        this._addPlaces(item.menu, 'network');
        if (item.menu.numMenuItems > 0) this.addMenuItem(item);
        item = new PopupMenu.PopupSubMenuMenuItem(_("Removable Devices"), true);
        item.icon.icon_name = 'media-removable';
        this._addPlaces(item.menu, 'device');
        if (item.menu.numMenuItems > 0) this.addMenuItem(item);
        this.parent();
    }
});

const PlacesMenuButton = new Lang.Class({
    Name: 'PlacesMenuButton',
    Extends: PanelMenu.Button,

    _init() {
        this.parent(1.0, _("Places"), false);
        this.menu.actor.add_style_class_name('dschaerf-menu');
        this._hbox = new St.BoxLayout({ style_class: 'panel-status-menu-box' });
        this._icon = new St.Icon({ icon_name: 'folder-symbolic', style_class: 'system-status-icon' });
        this._hbox.add_child(this._icon);
        this._label = new St.Label({ text: _("Places"), y_expand: true, y_align: Clutter.ActorAlign.CENTER });
        this._hbox.add_child(this._label);
        this._hbox.add_child(PopupMenu.arrowIcon(St.Side.BOTTOM));
        this.actor.add_actor(this._hbox);
        this._itemBookmarks = new PopupPlacesBookmarksMenuSection();
        this.menu.addMenuItem(this._itemBookmarks);
        this.menu.addMenuItem(new PopupMenu.PopupSeparatorMenuItem());
        this._itemVolumes = new PopupPlacesVolumesMenuSection();
        this.menu.addMenuItem(this._itemVolumes);
        this.menu.addMenuItem(new PopupMenu.PopupSeparatorMenuItem());
        this._itemRecent = new PopupPlacesRecentMenuSection();
        this.menu.addMenuItem(this._itemRecent);
        this._settings = Convenience.getSettings();
        this._settingsChangedId = this._settings.connect('changed::%s'.format(KEY_SHOW_PLACES), this._onChangedShowPlaces.bind(this));
        this._onChangedShowPlaces();
    },

    _onChangedShowPlaces() {
        let showPlaces = this._settings.get_enum(KEY_SHOW_PLACES);
        this._icon.visible = showPlaces !== PANELMENU_BUTTON_TEXT;
        this._label.visible = showPlaces !== PANELMENU_BUTTON_ICON;
    },

    destroy() {
        this._settings.disconnect(this._settingsChangedId);
        this.parent();
    }
});

function getExtension() {
    return Main.panel.addToStatusArea('PanelManagerPlacesMenu', new PlacesMenuButton(), 1, 'left');
}
