/* prefs.js
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * SPDX-License-Identifier: GPL-2.0-or-later
 */

/* exported init buildPrefsWidget */

const Gio = imports.gi.Gio;
const GLib = imports.gi.GLib;
const GObject = imports.gi.GObject;
const Gtk = imports.gi.Gtk;

const ExtensionUtils = imports.misc.extensionUtils;

const Me = ExtensionUtils.getCurrentExtension();

const Convenience = Me.imports.convenience;

const _ = Convenience.getTranslations();

const SETTINGS_GENERALLY = [
    {
        type: 'combo',
        key: 'show-activities',
        label: _("Activities button"),
        tooltip: _("Select the display mode for the Activities button in the top bar."),
        values: {
            'default': _("Default"),
            'icon': _("GNOME icon"),
            'none': _("Not show"),
        },
    },
    {
        type: 'switcher',
        key: 'always-show-universal-access-status',
        label: _("Always show the Universal Access status icon"),
        tooltip: _("This key overrides the automatic hiding of the Universal Access status icon when no accessibility features are enabled."),
        schema: 'org.gnome.desktop.a11y',
    },
    {
        type: 'switcher',
        key: 'always-show-input-sources-status',
        label: _("Always show the Input Sources status icon"),
        tooltip: _("This key overrides the automatic hiding of the Input Sources status icon when when only one Input Sources are enabled."),
    },
    {
        type: 'switcher',
        key: 'show-dropdown-arrow',
        label: _("Dropdown arrow"),
        tooltip: _("Determines whether or not buttons appear in the top bar with a dropdown arrow."),
    },
    {
        type: 'separator',
    },
    {
        type: 'switcher',
        key: 'show-dash',
        label: _("Dash"),
        tooltip: _("Determines whether the Dash is displayed or not."),
    },
    {
        type: 'switcher',
        key: 'show-dash-applications',
        label: _("Show Applications icon on Dash"),
        tooltip: _("Determines whether the Show Applications icon in the Dash is displayed or not."),
    },
    {
        type: 'separator',
    },
    {
        type: 'switcher',
        key: 'center-new-windows',
        label: _("Place new windows in the center"),
        tooltip: _("When true, the new windows will always be put in the center of the active screen of the monitor."),
        schema: 'org.gnome.mutter',
    },
];

const SETTINGS_HOTCORNER = [
    {
        type: 'combo',
        key: 'show-hotcorner',
        label: _("HotCorner"),
        tooltip: _("Select the display mode for HotCorner in the top bar."),
        values: {
            'default': _("Default"),
            'alternative': _("Alternative"),
            'disabled': _("Disabled"),
        },
    },
    {
        type: 'separator',
    },
    {
        type: 'switcher',
        key: 'show-hotcorner-left',
        label: _("Alternative HotCorner left"),
        tooltip: _("Determines if the alternative left HotCorner is enabled or not."),
    },
    {
        type: 'switcher',
        key: 'show-hotcorner-right',
        label: _("Alternative HotCorner right"),
        tooltip: _("Determines if the alternative right HotCorner is enabled or not."),
    },
    {
        type: 'slider',
        key: 'hotcorner-sensitivity',
        label: _("Response sensitivity for alternative HotCorner"),
        tooltip: _("Setting the response sensitivity for alternative HotCorner."),
        minimal: 1,
        maximum: 10,
        step: 1,
    },
];

const SETTINGS_APPLICATIONS = [
    {
        type: 'combo',
        key: 'show-applications',
        label: _("Applications button"),
        tooltip: _("Select the display mode for the Applications button in the top bar."),
        values: {
            'all' : _("Icon and Text"),
            'icon': _("Icon"),
            'text': _("Text"),
        },
    },
    {
        type: 'separator',
    },
    {
        type: 'label',
        label: _("To specify which applications will be displayed\nthe GNOME menu editor (alacarte) is required."),
    },
];
if (GLib.file_test('/usr/bin/alacarte', GLib.FileTest.EXISTS)) SETTINGS_APPLICATIONS.splice(2, 1, {
    type : 'button',
    label: _("Change which applications are shown"),
    tooltip: _("Use the GNOME menu editor to specify which applications will be displayed."),
    value: _("Main Menu"),
    app: 'alacarte',
});

const SETTINGS_PLACES = [
    {
        type: 'combo',
        key: 'show-places',
        label: _("Places button"),
        tooltip: _("Select the display mode for the Places button in the top bar."),
        values: {
            'all': _("Icon and Text"),
            'icon': _("Icon"),
            'text': _("Text"),
        },
    },
    {
        type: 'combo',
        key: 'bookmarks-position',
        label: _("Position of Bookmarks"),
        tooltip: _("Select the position for the bookmarks in the places menu. Custom bookmarks are displayed in a submenu."),
        values: {
             'mainmenu': _("Main Menu"),
             'submenu': _("Submenu"),
        },
    },
    {
        type: 'separator',
    },
    {
        type: 'switcher',
        key: 'show-desktop-icons',
        label: _("Have file manager handle the desktop"),
        tooltip: _("If set to true, then file manager will draw the icons on the desktop."),
        schema: 'org.gnome.desktop.background',
    },
    {
        type: 'separator',
    },
    {
        type: 'spinner',
        key: 'max-trash-entries',
        label: _("Maximum entries for Last Trashed"),
        tooltip: _("This number set the maximum entries for Last Trashed in the Trash menu."),
        minimal: 0,
        maximum: 100,
        step: 5,
    },
    {
        type: 'label',
        label: _("0 = All entries in Last Trashed are displayed\nNote: Too mutch can affect the display of the entries"),
    },
];

const SETTINGS_FAVORITES = [
    {
        type: 'combo',
        key: 'show-favorites',
        label: _("Favorites button"),
        tooltip: _("Select the display mode for the Favorites button in the top bar."),
        values: {
            'all': _("Icon and Text"),
            'icon': _("Icon"),
            'text': _("Text"),
        },
    },
    {
        type: 'combo',
        key: 'favorites-scope',
        label: _("Menu of the Favorites button"),
        tooltip: _("Change which favorites are shown in the Favorites button menu."),
        values: {
            'own': _("Only own Favorites"),
            'all': _("All Favorites"),
        },
    },
    {
        type: 'favorites',
        key: 'favorite-apps',
        label: _("Own Favorites"),
    },
];

const SETTINGS_DATEMENU = [
    {   type:    'switcher',
        key:     'show-events-icon',
        label:   _("Show day events icon"),
        tooltip: _("If true, display day events icon, in addition to time."),
    },
    {   type:    'switcher',
        key:     'clock-show-date',
        label:   _("Show date in clock"),
        tooltip: _("If true, display date in the clock, in addition to time."),
        schema:  'org.gnome.desktop.interface',
    },
    {   type:    'switcher',
        key:     'clock-show-seconds',
        label:   _("Whether the clock shows seconds"),
        tooltip: _("If true, display seconds in the clock."),
        schema:  'org.gnome.desktop.interface',
    },
    {   type:    'switcher',
        key:     'clock-show-weekday',
        label:   _("Show weekday in clock"),
        tooltip: _("If true, display weekday in the clock, in addition to time."),
        schema:  'org.gnome.desktop.interface',
    },
    {
        type: 'separator',
    },
    {   type:    'switcher',
        key:     'show-weekdate',
        label:   _("Show the week date in the calendar"),
        tooltip: _("If true, display the ISO week date in the calendar."),
        schema:  'org.gnome.desktop.calendar',
    },
];

const SETTINGS_SYSTEMMENU = [
    {
        type: 'combo',
        key: 'show-systemmenu',
        label: _("System menu button"),
        tooltip: _("Select the display mode for the System menu button in the top bar."),
        values: {
            'default': _("Default"),
            'icon': _("Shutdown icon")
        }
    },
    {   type:    'switcher',
        key:     'show-battery-percentage',
        label:   _("Show battery percentage"),
        tooltip: _("If true, display battery percentage in the status menu, in addition to the icon."),
        schema: 'org.gnome.desktop.interface',
    },
    {
        type: 'separator',
    },
    {
        type: 'switcher',
        key: 'always-show-audio-input',
        label: _("Always show the Audio Input widget"),
        tooltip: _("This key overrides the automatic hiding of the Audio Input widget when none application is recording audio."),
    },
    {
        type: 'separator',
    },
];

function checkAppId(appId) {
    if (appId.indexOf('.desktop') < 0) return false;
    if (GLib.file_test(appId, GLib.FileTest.EXISTS) ||
        GLib.file_test('/usr/share/applications/'+appId, GLib.FileTest.EXISTS) ||
        GLib.file_test(GLib.get_home_dir()+'/.local/share/applications/'+appId, GLib.FileTest.EXISTS) ||
        Glib.find_program_in_path(appId)) return true;
    return false;
}

const PanelManagerPrefsWidget = GObject.registerClass(
class PanelManagerPrefsWidget extends Gtk.Notebook {
    _init() {
        super._init({ scrollable: true, margin: 15, width_request: 720, height_request: 480 });
        this._createPrefsWidget(SETTINGS_GENERALLY, _("Generally"));
        this._createPrefsWidget(SETTINGS_HOTCORNER, _("HotCorner"));
        this._createPrefsWidget(SETTINGS_APPLICATIONS, _("Applications"));
        this._createPrefsWidget(SETTINGS_PLACES, _("Places"));
        this._createPrefsWidget(SETTINGS_FAVORITES, _("Favorites"));
        this._createPrefsWidget(SETTINGS_DATEMENU, _("Date and Time"));
        this._createPrefsWidget(SETTINGS_SYSTEMMENU, _("System menu"));
        this.show_all();
    }

    _createDialog() {
        let dialog = new Gtk.Dialog({ title: _("Add Favorite"), modal: true, resizable: false, transient_for: this.get_toplevel(), use_header_bar: true });
        dialog.add_button(_("Cancel"), Gtk.ResponseType.CANCEL);
        return dialog;
    }

    _createPrefsWidget(prefs, text) {
        let box = new Gtk.Box({ orientation: Gtk.Orientation.VERTICAL, margin: 15, spacing: 15 });
        prefs.forEach(pref => {
            let widget = this._createWidget(pref);
            if (widget) box.add(widget);
        });
        this.append_page(box, new Gtk.Label({ label: text }));
    }

    _createWidget(pref) {
        let settings = Convenience.getSettings(pref.schema ? pref.schema : null);
        let widget = null;
        switch (pref.type) {
        case 'favorites':
            widget = new Gtk.Box({ orientation: Gtk.Orientation.VERTICAL });
            let frame = new Gtk.Frame();
            widget.add(frame);
            let scrolledWindow = new Gtk.ScrolledWindow({ expand: true });
            scrolledWindow.set_policy(Gtk.PolicyType.NEVER, Gtk.PolicyType.AUTOMATIC);
            frame.add(scrolledWindow);
            let listStore = new Gtk.ListStore();
            listStore.set_column_types([ GObject.TYPE_STRING, GObject.TYPE_STRING ]);
            let treeView = new Gtk.TreeView({ expand: true, model: listStore });
            scrolledWindow.add(treeView);
            treeView.get_selection().set_mode(Gtk.SelectionMode.SINGLE);
            let columnFavorites = new Gtk.TreeViewColumn({ expand: true, sort_column_id: 1, title: _(pref.label) });
            let textRenderer = new Gtk.CellRendererText();
            columnFavorites.pack_start(textRenderer, true);
            columnFavorites.add_attribute(textRenderer, 'text', 1);
            treeView.insert_column(columnFavorites, 0);
            let toolbar = new Gtk.Toolbar();
            toolbar.get_style_context().add_class(Gtk.STYLE_CLASS_INLINE_TOOLBAR);
            widget.add(toolbar);
            let addButton = new Gtk.ToolButton({ icon_name: 'list-add', label: _("Select application"), is_important: true });
            addButton.connect('clicked', () => {
                this._favoriteSearch(settings, pref.key);
            });
            toolbar.add(addButton);
            let inputButton = new Gtk.ToolButton({ icon_name: 'list-add', label: _("Desktop file ID"), is_important: true });
            inputButton.connect('clicked', () => {
                this._favoriteInput(settings, pref.key);
            });
            toolbar.add(inputButton);
            let delButton = new Gtk.ToolButton({ icon_name: 'list-remove', label: _("Remove Favorite"), is_important: true });
            delButton.connect('clicked', () => {
                let [ any, model, iter ] = treeView.get_selection().get_selected();
                strArray = settings.get_strv(pref.key);
                if (any && strArray.length > 0) {
                    let item = listStore.get_value(iter, 0);
                    let index = strArray.map(i => {
                        return i;
                    }).indexOf(item);
                    if (index < 0) return;
                    strArray.splice(index, 1);
                    settings.set_strv(pref.key, strArray);
                }
            });
            toolbar.add(delButton);
            let strArray = [];
            let treeViewSelection = treeView.get_selection();
            let updateListStore = () => {
                strArray = settings.get_strv(pref.key);
                strArray.sort((a, b) => {
                    try { a = Gio.DesktopAppInfo.new(a).get_name(); } catch(e) {}
                    try { b = Gio.DesktopAppInfo.new(b).get_name(); } catch(e) {}
                    return a.toLowerCase() > b.toLowerCase();
                });
                listStore.clear();
                strArray.forEach(app => {
                    if (Gio.DesktopAppInfo.new(app)) {
                        let item = Gio.DesktopAppInfo.new(app).get_name();
                        listStore.set(listStore.append(), [ 0, 1 ], [ app, item ]);
                    }
                });
                treeViewSelection.select_path(Gtk.TreePath.new_from_string(String(0)));
            }
            settings.connect('changed::'+pref.key, updateListStore);
            updateListStore();
            break;
        case 'label':
            widget = new Gtk.Label({ label: _(pref.label), xalign: 0 });
            break;
        case 'separator':
            widget = new Gtk.Separator();
            break;
        default:
            widget = new Gtk.Box({ orientation: Gtk.Orientation.HORIZONTAL });
            widget.pack_start(new Gtk.Label({ label: _(pref.label), xalign: 0 }), true, true, 0);
            let prefAction;
            switch (pref.type) {
            case 'button':
                prefAction = new Gtk.Button({ label: _(pref.value) });
                prefAction.connect('pressed', () => {
                    let [success, argv] = GLib.shell_parse_argv(pref.app);
                    GLib.spawn_async(null, argv, null, GLib.SpawnFlags.SEARCH_PATH, null);
                });
                break;
            case 'combo':
                prefAction = new Gtk.ComboBoxText();
                for (let value in pref.values) {
                    prefAction.append(value, _(pref.values[value]));
                }
                prefAction.connect('changed', () => {
                    settings.set_string(pref.key, prefAction.get_active_id());
                });
                prefAction.set_active_id(settings.get_string(pref.key));
                settings.connect('changed::%s'.format(pref.key), () => {
                    prefAction.set_active_id(settings.get_string(pref.key));
                });
                break;
            case 'slider':
                prefAction = new Gtk.Scale({ orientation: Gtk.Orientation.HORIZONTAL, valuePos: Gtk.PositionType.LEFT });
                prefAction.set_digits(0);
                prefAction.set_increments(pref.step, pref.step);
                prefAction.set_range(pref.minimal, pref.maximum);
                prefAction.set_value(settings.get_int(pref.key));
                prefAction._timeoutId = 0;
                prefAction.connect('value-changed', () => {
                    if (prefAction._timeoutId > 0) return;
                    prefAction._timeoutId = GLib.timeout_add(GLib.PRIORITY_DEFAULT, 100, () => {
                        prefAction._timeoutId = 0;
                        settings.set_int(pref.key, prefAction.get_value());
                        return GLib.SOURCE_REMOVE;
                    });
                });
                settings.connect('changed::%s'.format(pref.key), () => {
                    prefAction.set_value(settings.get_int(pref.key));
                });
                break;
            case 'spinner':
                prefAction = Gtk.SpinButton.new_with_range(pref.minimal, pref.maximum, pref.step);
                prefAction.set_value(settings.get_int(pref.key));
                prefAction.set_digits(0);
                prefAction.connect('value-changed', () => {
                    if (prefAction._timeoutId > 0) return;
                    prefAction._timeoutId = GLib.timeout_add(GLib.PRIORITY_DEFAULT, 100, () => {
                        prefAction._timeoutId = 0;
                        settings.set_int(pref.key, prefAction.get_value());
                        return GLib.SOURCE_REMOVE;
                    });
                });
                settings.connect('changed::%s'.format(pref.key), () => {
                    prefAction.set_value(settings.get_int(pref.key));
                });
                break;
            case 'switcher':
                prefAction = new Gtk.Switch({ active: settings.get_boolean(pref.key) });
                prefAction.connect('notify::active', () => {
                    settings.set_boolean(pref.key, prefAction.get_active());
                });
                settings.connect('changed::%s'.format(pref.key), () => {
                    prefAction.set_active(settings.get_boolean(pref.key));
                });
                break;
            default: // Nothing to do
            }
            if (pref.type !== 'spinner' && pref.type !== 'switcher') prefAction.width_request = 240;
            if (pref.tooltip) prefAction.set_tooltip_text(_(pref.tooltip));
            widget.pack_end(prefAction, false, false, 0);
        }
        return widget;
    }

    _favoriteInput(schema, key) {
        let dialog = this._createDialog();
        let addButton = dialog.add_button(_("Add"), Gtk.ResponseType.OK);
        addButton.sensitive = 0;
        dialog.set_default_response(Gtk.ResponseType.OK);
        let grid = new Gtk.Grid({ column_spacing: 15, margin: 15 });
        dialog.get_content_area().add(grid);
        grid.attach(new Gtk.Label({ label: _("Desktop file ID:"), xalign: 0 }), 0, 0, 1, 1);
        let entry = new Gtk.Entry({ activates_default: true, expand: true });
        entry.connect('changed', () => {
            addButton.sensitive = 0;
            let appId = entry.get_text();
            if (checkAppId(appId)) {
                let strArray = schema.get_strv(key);
                if (strArray.indexOf(appId) >= 0) return;
                addButton.sensitive = 1;
            }
        });
        grid.attach(entry, 1, 0, 1, 1);
        dialog.connect('response', (dialog, response_id) => {
            if (response_id === Gtk.ResponseType.OK) {
                let appId = entry.get_text();
                let strArray = schema.get_strv(key);
                strArray.push(appId);
                schema.set_strv(key, strArray);
            }
            dialog.destroy();
        });
        dialog.show_all();
    }

    _favoriteSearch(schema, key) {
        let dialog = this._createDialog();
        let addButton = dialog.add_button(_("Add"), Gtk.ResponseType.OK);
        addButton.sensitive = 0;
        dialog.set_default_response(Gtk.ResponseType.OK);
        let grid = new Gtk.Grid({ margin: 15 });
        dialog.get_content_area().add(grid);
        let appChooser = new Gtk.AppChooserWidget({ expand: true });
        appChooser.set_default_text('');
        appChooser.set_show_all(true);
        appChooser.connect('application-selected', () => {
            addButton.sensitive = 0;
            let appInfo = appChooser.get_app_info();
            if (!appInfo) return;
            let strArray = schema.get_strv(key);
            if (strArray.indexOf(appInfo.get_id()) >= 0) return;
            addButton.sensitive = 1;
        });
        grid.attach(appChooser, 0, 0, 1, 1);
        dialog.connect('response', (dialog, response_id) => {
            if (response_id === Gtk.ResponseType.OK) {
                let appInfo = appChooser.get_app_info();
                let strArray = schema.get_strv(key);
                strArray.push(appInfo.get_id());
                schema.set_strv(key, strArray);
            }
            dialog.destroy();
        });
        dialog.show_all();
    }
});

function init(metadata) {
    Convenience.initTranslations();
}

function buildPrefsWidget() {
    return new PanelManagerPrefsWidget();
}
