/* recycleBin.js
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * SPDX-License-Identifier: GPL-2.0-or-later
 */

/* exported getExtension */

const Lang = imports.lang;

const Clutter = imports.gi.Clutter;
const St = imports.gi.St;

const ExtensionUtils = imports.misc.extensionUtils;

const Main = imports.ui.main;
const PanelMenu = imports.ui.panelMenu;
const PopupMenu = imports.ui.popupMenu;

const Me = ExtensionUtils.getCurrentExtension();

const Convenience = Me.imports.convenience;
const PlacesManager = Me.imports.placesManager;
const UserDialog = Me.imports.userDialog;

const _ = Convenience.getTranslations();

const KEY_MAX_TRASH_ENTRIES = 'max-trash-entries';

const PopupTrashMenuItem = new Lang.Class({
    Name: 'PopupTrashMenuItem',
    Extends: PopupMenu.PopupImageMenuItem,

    _init(info) {
        this.parent(info.name, info.icon);
        this.label.set({ x_expand: true, y_expand: true, y_align: Clutter.ActorAlign.CENTER });
        if (info.can_empty) {
            this._addButton({
                icon: 'edit-clear-all-symbolic',
                content: 'emptyTrash',
                value: null,
                signal: 'empty',
                callback: () => { info.empty(); },
            });
        } else if (info.can_remove) {
            this._addButton({
                icon: 'edit-clear-symbolic',
                content: 'deleteTrashEntry',
                value: info.name,
                signal: 'remove',
                callback: () => { info.remove(); },
            });
        }
        this._info = info;
    },

    _addButton(params) {
        this.button = new St.Button({ style_class: 'system-menu-action popup-menu-button' });
        this.button.set_child(new St.Icon({ icon_name: params.icon }));
        this.button.connect('clicked', event => {
            let dialog = new UserDialog.ModalUserDialog(params.content, params.value);
            dialog.connect(params.signal, params.callback);
            Main.overview.hide();
            dialog.open();
            this.emit('activate', event);
        });
        this.actor.add_child(this.button);
    },

    activate(event) {
        Main.overview.hide();
        this._info.launch(event.get_time());
        this.parent(event);
    }
});

const PopupTrashSubMenuMenuItem = new Lang.Class({
    Name: 'PopupTrashSubMenuMenuItem',
    Extends: PopupMenu.PopupSubMenuMenuItem,

    _init(text, icon) {
        this.parent(text, true);
        this.icon.icon_name = icon;
        this._statusLabel = new St.Label({ y_align: Clutter.ActorAlign.CENTER });
        let children = this.actor.get_n_children();
        this.actor.insert_child_at_index(this._statusLabel, children - 1);
    },

    setStatus(text) {
        this._statusLabel.set_text(text);
    }
});

const TrashMenuButton = new Lang.Class({
    Name: 'TrashMenuButton',
    Extends: PanelMenu.Button,

    _init() {
        this.parent(0.0, _("Trash"), false);
        this.menu.actor.add_style_class_name('dschaerf-menu');
        this._hbox = new St.BoxLayout({ style_class: 'panel-status-menu-box' });
        this._icon = new St.Icon({ icon_name: 'user-trash-symbolic', style_class: 'system-status-icon' });
        this._hbox.add_child(this._icon);
        this._hbox.add_child(PopupMenu.arrowIcon(St.Side.BOTTOM));
        this.actor.add_child(this._hbox);
        this._placesManager = new PlacesManager.PlacesTrashManager();
        this._placesManagerChangedId = this._placesManager.connect('updated', this._reload.bind(this));
        this._settings = Convenience.getSettings();
        this._settingsChangedId = this._settings.connect('changed::%s'.format(KEY_MAX_TRASH_ENTRIES), this._reload.bind(this));
        this._reload();
    },

    _addPlaces(menu, kind, maximum) {
        let places = this._placesManager.get(kind);
        for (let i = 0; i < places.length; i++) {
            if (maximum && i >= maximum) break;
            menu.addMenuItem(new PopupTrashMenuItem(places[i]));
        }
    },

    _reload() {
        this.menu.removeAll();
        if (this._placesManager.size > 0) {
            let item;
            this._addPlaces(this.menu, 'trash');
            item = new PopupTrashSubMenuMenuItem(_("Last Trashed"), 'user-trash-full');
            let maximum = this._settings.get_int(KEY_MAX_TRASH_ENTRIES);
            this._addPlaces(item.menu, 'trash-file', maximum > 0 ? maximum : null);
            if (maximum > 0 && maximum < this._placesManager.size) item.setStatus('(%s)'.format(maximum));
            this.menu.addMenuItem(item);
            this.menu.addMenuItem(new PopupMenu.PopupSeparatorMenuItem());
            item = new PopupMenu.PopupSwitchMenuItem(_("Numbers of objects in the Trash"), false);
            item.setStatus('%s'.format(this._placesManager.size));
            this.menu.addMenuItem(item);
        }
        this.actor.visible = this._placesManager.size > 0;
    },

    destroy() {
        this._settings.disconnect(this._settingsChangedId);
        this._placesManager.disconnect(this._placesManagerChangedId);
        this._placesManager.destroy();
        this.parent();
    }
});

function getExtension() {
    let pos = Main.sessionMode.panel.right.indexOf('aggregateMenu');
    return Main.panel.addToStatusArea('PanelManagerTrashMenu', new TrashMenuButton(), pos, 'right');
}
