/* systemMenu.js
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * SPDX-License-Identifier: GPL-2.0-or-later
 */

/* exported getExtension */

const Lang = imports.lang;
const Mainloop = imports.mainloop;

const Gio = imports.gi.Gio;
const GLib = imports.gi.GLib;
const Gvc = imports.gi.Gvc;
const Meta = imports.gi.Meta;
const Shell = imports.gi.Shell;
const St = imports.gi.St;

const ExtensionUtils = imports.misc.extensionUtils;
const LoginManager = imports.misc.loginManager;
const Util = imports.misc.util;

const BoxPointer = imports.ui.boxpointer;
const Main = imports.ui.main;
const PanelMenu = imports.ui.panelMenu;
const PopupMenu = imports.ui.popupMenu;

const StatusSystem = imports.ui.status.system;
const StatusVolume = imports.ui.status.volume;

const Me = ExtensionUtils.getCurrentExtension();

const Convenience = Me.imports.convenience;
const UserDialog = Me.imports.userDialog;

const _ = Convenience.getTranslations();

const BASIC_SETTINGS = [
    'ca.desrt.dconf-editor.desktop',        // dconf-editor
    'gconf-editor.desktop',                 // gconf-editor
    'gnome-control-center.desktop',         // gnome-control-center
    'gnome-shell-extension-prefs.desktop',  // gnome-shell
    'org.gnome.tweaks.desktop',             // gnome-tweaks
    'system-config-printer.desktop',        // system-config-printer
    'synaptic.desktop',                     // synaptic
    'users.desktop'                         // gnome-system-tools
]

const GNOME_TERMINAL = 'org.gnome.Terminal.desktop';

const KEY_ALWAYS_SHOW_AUDIO_INPUT = 'always-show-audio-input';
const KEY_DISABLE_RESTART_BUTTONS = 'disable-restart-buttons';
const KEY_SHOW_SYSTEMMENU = 'show-systemmenu';

const SYSTEMMENU_DEFAULT = 0;
const SYSTEMMENU_ICON = 1;

const PopupVolumeSubMenuMenuItem = new Lang.Class({
    Name: 'PopupVolumeSubMenuMenuItem',
    Extends: PopupMenu.PopupSubMenuMenuItem,

    _init(deviceType) {
        this.parent('...', true);
        this._controlChangedId = [
            Main.panel.statusArea.aggregateMenu._volume._control.connect('state-changed', this._updateDevices.bind(this)),
            Main.panel.statusArea.aggregateMenu._volume._control.connect('%s-added'.format(deviceType), this._updateDevices.bind(this)),
            Main.panel.statusArea.aggregateMenu._volume._control.connect('%s-removed'.format(deviceType), this._updateDevices.bind(this)),
            Main.panel.statusArea.aggregateMenu._volume._control.connect('active-%s-update'.format(deviceType), this._setActiveDevice.bind(this)),
        ];
        this._updateDevices();
    },

    _addDevice(id) {
        let uidevice = this.getDeviceById(id);
        if (!uidevice || !uidevice.port_available) return null;
        let deviceInfo = new Object();
        deviceInfo.uidevice = uidevice;
        deviceInfo.text = uidevice.description;
        if (uidevice.origin !== '') deviceInfo.text += ' - '+uidevice.origin;
        deviceInfo.item = this.menu.addAction(deviceInfo.text, () => {
            this.changeDevice(uidevice);
        });
        this._devices[id] = deviceInfo;
        return uidevice;
    },

    _setActiveDevice(control, id) {
        let deviceInfo = this._devices[id];
        if (deviceInfo && deviceInfo !== this._activeDevice) {
            if (this._activeDevice) this._activeDevice.item.setOrnament(PopupMenu.Ornament.NONE);
            this._activeDevice = deviceInfo;
            deviceInfo.item.setOrnament(PopupMenu.Ornament.CHECK);
            this.label.text = deviceInfo.text;
            let icon = deviceInfo.uidevice.get_icon_name();
            if (icon === null || icon.trim() === '') icon = this.getDefaultIcon();
            this.icon.icon_name = icon;
        }
    },

    _updateDevices() {
        this.menu.removeAll();
        this._devices = {};
        if (Main.panel.statusArea.aggregateMenu._volume._control.get_state() === Gvc.MixerControlState.READY) {
            let defaultStream = this.getDefaultStream();
            let numId = new Gvc.MixerUIDevice().get_id();
            for (let id = 0; id <= numId; id++) {
                let uidevice = this._addDevice(id);
                if (uidevice) {
                    let stream = Main.panel.statusArea.aggregateMenu._volume._control.get_stream_from_device(uidevice);
                    if (stream) {
                        let stream_port = stream.get_port();
                        if (stream_port && stream === defaultStream && stream_port.port === uidevice.get_port()) this._setActiveDevice(Main.panel.statusArea.aggregateMenu._volume._control, id);
                    }
                }
            }
        }
        this.setVisibility(true);
    },

    setVisibility(visible) {
        this.actor.visible = this.menu.numMenuItems > 1 && visible;
    },

    destroy() {
        this._controlChangedId.forEach(id => {
            Main.panel.statusArea.aggregateMenu._volume._control.disconnect(id);
        });
        this.parent();
    }
});

const PopupVolumeInputSubMenuMenuItem = new Lang.Class({
    Name: 'PopupVolumeInputSubMenuMenuItem',
    Extends: PopupVolumeSubMenuMenuItem,

    _init() {
        this.parent('input');
    },

    changeDevice(uidevice) {
        Main.panel.statusArea.aggregateMenu._volume._control.change_input(uidevice);
    },

    getDefaultIcon() {
        return 'audio-input-microphone';
    },

    getDefaultStream() {
        return Main.panel.statusArea.aggregateMenu._volume._control.get_default_source();
    },

    getDeviceById(id) {
        return Main.panel.statusArea.aggregateMenu._volume._control.lookup_input_id(id);
    }
});

const PopupVolumeOutputSubMenuMenuItem = new Lang.Class({
    Name: 'PopupVolumeOutputSubMenuMenuItem',
    Extends: PopupVolumeSubMenuMenuItem,

    _init() {
        this.parent('output');
    },

    changeDevice(uidevice) {
        Main.panel.statusArea.aggregateMenu._volume._control.change_output(uidevice);
    },

    getDefaultIcon() {
        return 'audio-card';
    },

    getDefaultStream() {
        return Main.panel.statusArea.aggregateMenu._volume._control.get_default_sink();
    },

    getDeviceById(id) {
        return Main.panel.statusArea.aggregateMenu._volume._control.lookup_output_id(id);
    }
});

const VolumeIndicator = new Lang.Class({
    Name: 'VolumeIndicator',

    _init() {
        this._secondaryIndicator = Main.panel.statusArea.aggregateMenu._volume._addIndicator();
        this._itemVolumeOutput = new PopupVolumeOutputSubMenuMenuItem();
        Main.panel.statusArea.aggregateMenu._volume._volumeMenu.addMenuItem(this._itemVolumeOutput, 1);
        this._itemVolumeInput = new PopupVolumeInputSubMenuMenuItem();
        Main.panel.statusArea.aggregateMenu._volume._volumeMenu.addMenuItem(this._itemVolumeInput, 3);
        Main.panel.statusArea.aggregateMenu._volume._volumeMenu._input._maybeShowInput = function() {
            let alwaysShow = Convenience.getSettings().get_boolean(KEY_ALWAYS_SHOW_AUDIO_INPUT);
            if (alwaysShow) {
                this._showInput = true;
                this._updateVisibility();
            } else {
                StatusVolume.InputStreamSlider.prototype._maybeShowInput.call(this);
            }
        };
        Main.panel.statusArea.aggregateMenu._volume._volumeMenu._input._shouldBeVisible = function() {
            let alwaysShow = Convenience.getSettings().get_boolean(KEY_ALWAYS_SHOW_AUDIO_INPUT);
            if (alwaysShow) {
                return this._stream !== null;
            } else {
                StatusVolume.InputStreamSlider.prototype._shouldBeVisible;
            }
        };
        this._iconChangedId = Main.panel.statusArea.aggregateMenu._volume._volumeMenu.connect('icon-changed', this._onIconChanged.bind(this));
        this._streamUpdatedId = Main.panel.statusArea.aggregateMenu._volume._volumeMenu._input.connect('stream-updated', this._onIconChanged.bind(this));
        this._settings = Convenience.getSettings();
        this._settingsChangedId = this._settings.connect('changed::%s'.format(KEY_ALWAYS_SHOW_AUDIO_INPUT), this._onAlwaysShowAudioInput.bind(this));
        if (Main.layoutManager._startingUp) {
            this._startupComplete = Main.layoutManager.connect('startup-complete', () => {
                this._onIconChanged();
                this._onAlwaysShowAudioInput();
                Main.layoutManager.disconnect(this._startupComplete);
            });
        } else {
            this._onIconChanged();
            this._onAlwaysShowAudioInput();
        }
    },

    _getIcon() {
        if (!Main.panel.statusArea.aggregateMenu._volume._volumeMenu._input._stream) return null;
        if (Main.panel.statusArea.aggregateMenu._volume._volumeMenu._input._stream.is_muted || Main.panel.statusArea.aggregateMenu._volume._volumeMenu._input._stream.volume <= 0) {
            return 'microphone-sensitivity-muted-symbolic';
        } else {
            let n = Math.floor(3 * Main.panel.statusArea.aggregateMenu._volume._volumeMenu._input._stream.volume / Main.panel.statusArea.aggregateMenu._volume._control.get_vol_max_norm()) + 1;
            if (n < 2) return 'microphone-sensitivity-low-symbolic';
            if (n >= 3) return 'microphone-sensitivity-high-symbolic';
            return 'microphone-sensitivity-medium-symbolic';
        }
    },

    _onAlwaysShowAudioInput() {
        this._maybeShowInput();
        let showInput = Main.panel.statusArea.aggregateMenu._volume._volumeMenu._input._showInput;
        this._secondaryIndicator.visible = showInput;
        this._itemVolumeInput.setVisibility(showInput);
    },

    _onIconChanged() {
        let icon = this._getIcon();
        if (icon !== null) {
            this._secondaryIndicator.icon_name = icon;
            this._secondaryIndicator.show();
        } else {
            this._secondaryIndicator.hide();
        }
        Main.panel.statusArea.aggregateMenu._volume.indicators.visible = Main.panel.statusArea.aggregateMenu._volume.indicators.visible || this._secondaryIndicator.visible;
    },

    _maybeShowInput() {
        Main.panel.statusArea.aggregateMenu._volume._volumeMenu._input._maybeShowInput.call(Main.panel.statusArea.aggregateMenu._volume._volumeMenu._input);
    },

    destroy() {
        this._settings.disconnect(this._settingsChangedId);
        Main.panel.statusArea.aggregateMenu._volume._volumeMenu._input.disconnect(this._streamUpdatedId);
        Main.panel.statusArea.aggregateMenu._volume._volumeMenu.disconnect(this._iconChangedId);
        Main.panel.statusArea.aggregateMenu._volume._volumeMenu._input._shouldBeVisible = StatusVolume.InputStreamSlider.prototype._shouldBeVisible;
        Main.panel.statusArea.aggregateMenu._volume._volumeMenu._input._maybeShowInput = StatusVolume.InputStreamSlider.prototype._maybeShowInput;
        this._itemVolumeInput.destroy();
        this._itemVolumeOutput.destroy();
        this._secondaryIndicator.destroy();
        this._maybeShowInput();
    }
});

const SystemIndicator = new Lang.Class({
    Name: 'SystemIndicator',

    _init() {
        this._itemBasicSettings = new PopupMenu.PopupSubMenuMenuItem(_("Basic Settings"), true);
        this._itemBasicSettings.icon.icon_name = 'preferences-system-symbolic';
        Main.panel.statusArea.aggregateMenu._system.menu.addMenuItem(this._itemBasicSettings, 0);
        let buttons = [];
        let children = Main.panel.statusArea.aggregateMenu._system._actionsItem.actor.get_children();
        for (let i = 1; i < children.length; i++) {
            Main.panel.statusArea.aggregateMenu._system._actionsItem.actor.remove_actor(children[i]);
            buttons.push(children[i]);
        }
        for (let i = 0; i < buttons.length; i++) {
            if (i === 1) {
                this._settingsAction = Main.panel.statusArea.aggregateMenu._system._createActionButton('preferences-other-symbolic', _("PanelManager Settings"));
                this._settingsAction.connect('clicked', this._onSettingsClicked.bind(this));
                Main.panel.statusArea.aggregateMenu._system._actionsItem.actor.add(this._settingsAction, { expand: true, x_fill: false });
                this._terminalAction = Main.panel.statusArea.aggregateMenu._system._createActionButton('utilities-terminal-symbolic', _("Terminal"));
                this._terminalAction.connect('clicked', this._onTerminalClicked.bind(this));
                Main.panel.statusArea.aggregateMenu._system._actionsItem.actor.add(this._terminalAction, { expand: true, x_fill: false });
                this._restartAction = Main.panel.statusArea.aggregateMenu._system._createActionButton('view-refresh-symbolic', _("Restart"));
                this._restartAction.connect('clicked', this._onRestartClicked.bind(this));
                Main.panel.statusArea.aggregateMenu._system._actionsItem.actor.add(this._restartAction, { expand: true, x_fill: false });
            } else if (i === buttons.length - 1) {
                this._hibernateAction = Main.panel.statusArea.aggregateMenu._system._createActionButton('document-save-symbolic', _("Hibernate"));
                this._hibernateAction.connect('clicked', this._onHibernateClicked.bind(this));
                this._hybridSleepAction = Main.panel.statusArea.aggregateMenu._system._createActionButton('document-save-as-symbolic', _("Hybrid-Sleep"));
                this._hybridSleepAction.connect('clicked', this._onHybridSleepClicked.bind(this));
                this._altSwitcher = new StatusSystem.AltSwitcher(this._hibernateAction, this._hybridSleepAction);
                Main.panel.statusArea.aggregateMenu._system._actionsItem.actor.add(this._altSwitcher.actor, { expand: true, x_fill: false });
            }
            Main.panel.statusArea.aggregateMenu._system._actionsItem.actor.add(buttons[i], { expand: true, x_fill: false });
        }
        Main.panel.statusArea.aggregateMenu._system._systemActions.forceUpdate();
        Main.panel.statusArea.aggregateMenu._system._systemActions._updateOrientationLock();
        Main.panel.statusArea.aggregateMenu._system._systemActions._updateOrientationLockIcon();
        Main.panel.statusArea.aggregateMenu._system._systemActions._sessionUpdated();
        this._appSystem = Shell.AppSystem.get_default();
        this._appSystemChangedId = this._appSystem.connect('installed-changed', this._onInstalledChanged.bind(this));
        this._onInstalledChanged();
        this._loginManager = LoginManager.getLoginManager();
        this._openStateChangedId = Main.panel.statusArea.aggregateMenu._system.menu.connect('open-state-changed', (menu, open) => {
            if (!open) return;
            Main.panel.statusArea.aggregateMenu._system._settingsAction.hide();
            this._terminalAction.visible = this._appSystem.lookup_app(GNOME_TERMINAL);
            this._restartAction.visible = Meta.is_wayland_compositor() ? false : true;
            this._updateHibernate();
            this._updateHybridSleep();
        });
    },

    _canHibernation(value, asyncCallback) {
        if (this._loginManager._proxy) {
            this._loginManager._proxy.call(value, null, Gio.DBusCallFlags.NONE, -1, null, (proxy, asyncResult) => {
                try {
                    let result = proxy.call_finish(asyncResult).deep_unpack();
                    let disableRestartButton = Main.panel.statusArea.aggregateMenu._system._systemActions._loginScreenSettings.get_boolean(KEY_DISABLE_RESTART_BUTTONS);
                    let disabled = Main.sessionMode.isLocked || (Main.sessionMode.isGreeter && disableRestartButton);
                    asyncCallback(result[0] !== 'no' && !disabled);
                } catch(e) {
                    asyncCallback(false);
                }
            });
        } else {
            Mainloop.idle_add(() => {
                asyncCallback(false);
                return false;
            });
        }
    },

    _onHibernateClicked(animate) {
        Main.panel.statusArea.aggregateMenu._system.menu.itemActivated(animate);
        if (this._hybridSleepAction.visible) {
            this._dialog = new UserDialog.ModalUserDialog('hibernation');
            this._dialog.connect('hybrid-sleep', () => {
                this._onHibernation('HybridSleep');
            });
        } else {
            this._dialog = new UserDialog.ModalUserDialog('hibernate');
        }
        this._dialog.connect('hibernate', () => {
            this._onHibernation('Hibernate');
        });
        Main.overview.hide();
        this._dialog.open();
    },

    _onHibernation(value) {
        if (this._loginManager._proxy) {
            this._loginManager._proxy.call(value, GLib.Variant.new('(b)', [true]), Gio.DBusCallFlags.NONE, -1, null, null);
        } else {
            this._loginManager.emit('prepare-for-sleep', true);
            this._loginManager.emit('prepare-for-sleep', false);
        }
    },

    _onHybridSleepClicked(animate) {
        Main.panel.statusArea.aggregateMenu._system.menu.itemActivated(animate);
        this._dialog = new UserDialog.ModalUserDialog('hybridSleep');
        this._dialog.connect('hybrid-sleep', () => {
            this._onHibernation('HybridSleep');
        });
        Main.overview.hide();
        this._dialog.open();
    },

    _onInstalledChanged() {
        this._itemBasicSettings.menu.removeAll();
        let appList = [];
        BASIC_SETTINGS.forEach(desktopFile => {
            let app = this._appSystem.lookup_app(desktopFile);
            if (app) appList.push(app);
        });
        appList.sort((a, b) => a.compare_by_name(b));
        appList.forEach(app => {
            let item = new PopupMenu.PopupMenuItem(app.get_name());
            item.connect('activate', () => {
                Main.panel.statusArea.aggregateMenu._system.menu.itemActivated(BoxPointer.PopupAnimation.NONE);
                Main.overview.hide();
                app.activate();
            });
            this._itemBasicSettings.menu.addMenuItem(item);
        });
    },

    _onRestartClicked(animate) {
        Main.panel.statusArea.aggregateMenu._system.menu.itemActivated(animate);
        Meta.restart(_("Restarting…"));
    },

    _onSettingsClicked(animate) {
        Main.panel.statusArea.aggregateMenu._system.menu.itemActivated(animate);
        Main.overview.hide();
        Util.spawnCommandLine('gnome-shell-extension-prefs '+Me.metadata['uuid']);
    },

    _onTerminalClicked(animate) {
        Main.panel.statusArea.aggregateMenu._system.menu.itemActivated(animate);
        Main.overview.hide();
        Util.spawnCommandLine('gnome-terminal');
    },

    _updateHibernate() {
        this._canHibernation("CanHibernate", (result) => {
            this._hibernateAction.visible = result;
        });
    },

    _updateHybridSleep() {
        this._canHibernation("CanHybridSleep", (result) => {
            this._hybridSleepAction.visible = result;
        });
    },

    destroy() {
        Main.panel.statusArea.aggregateMenu._system.menu.disconnect(this._openStateChangedId);
        Main.panel.statusArea.aggregateMenu._system._actionsItem.actor.remove_actor(this._altSwitcher.actor);
        Main.panel.statusArea.aggregateMenu._system._actionsItem.actor.remove_actor(this._restartAction);
        Main.panel.statusArea.aggregateMenu._system._actionsItem.actor.remove_actor(this._terminalAction);
        Main.panel.statusArea.aggregateMenu._system._actionsItem.actor.remove_actor(this._settingsAction);
        Main.panel.statusArea.aggregateMenu._system._settingsAction.show();
        this._appSystem.disconnect(this._appSystemChangedId);
        this._hybridSleepAction.destroy();
        this._hibernateAction.destroy();
        this._altSwitcher.actor.destroy();
        this._restartAction.destroy();
        this._terminalAction.destroy();
        this._settingsAction.destroy();
        this._itemBasicSettings.destroy();
    }
});

const SystemMenuButton = new Lang.Class({
    Name: 'SystemMenuButton',

    _init() {
        Main.panel.statusArea.aggregateMenu.actor.remove_actor(Main.panel.statusArea.aggregateMenu._indicators);
        this._hbox = new St.BoxLayout();
        this._hbox.add_child(Main.panel.statusArea.aggregateMenu._indicators);
        this._indicator = new St.BoxLayout({ style_class: 'panel-status-indicators-box' });
        this._indicator.add_child(new St.Icon({ icon_name: 'system-shutdown-symbolic', style_class: 'system-status-icon' }));
        this._indicator.add_child(PopupMenu.arrowIcon(St.Side.BOTTOM));
        this._hbox.add_child(this._indicator);
        Main.panel.statusArea.aggregateMenu.actor.add_child(this._hbox);
        this._settings = Convenience.getSettings();
        this._settingsChangedId = this._settings.connect('changed::%s'.format(KEY_SHOW_SYSTEMMENU), this._onChangedShowSystemMenu.bind(this));
        this._onChangedShowSystemMenu();
        this._volume = new VolumeIndicator();
        this._system = new SystemIndicator();
    },

    _onChangedShowSystemMenu() {
        let showSystemMenu = this._settings.get_enum(KEY_SHOW_SYSTEMMENU);
        Main.panel.statusArea.aggregateMenu._indicators.visible = showSystemMenu === SYSTEMMENU_DEFAULT;
        this._indicator.visible = showSystemMenu === SYSTEMMENU_ICON;
    },

    destroy() {
        this._system.destroy();
        this._volume.destroy();
        this._settings.disconnect(this._settingsChangedId);
        Main.panel.statusArea.aggregateMenu.actor.remove_actor(this._hbox);
        this._hbox.remove_actor(Main.panel.statusArea.aggregateMenu._indicators);
        this._hbox.destroy_all_children();
        this._hbox.destroy();
        Main.panel.statusArea.aggregateMenu.actor.add_child(Main.panel.statusArea.aggregateMenu._indicators);
        Main.panel.statusArea.aggregateMenu._indicators.show();
    }
});

function getExtension() {
    return new SystemMenuButton();
}
