/* userDialog.js
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * SPDX-License-Identifier: GPL-2.0-or-later
 */

/* exported ModalUserDialog */

const Lang = imports.lang;

const Clutter = imports.gi.Clutter;
const Gio = imports.gi.Gio;
const Pango = imports.gi.Pango;
const St = imports.gi.St;

const ExtensionUtils = imports.misc.extensionUtils;

const Dialog = imports.ui.dialog;
const ModalDialog = imports.ui.modalDialog;

const Me = ExtensionUtils.getCurrentExtension();

const Convenience = Me.imports.convenience;

const _ = Convenience.getTranslations();

const clearRecentDocumentsDialogContent = {
    icon: 'document-open-recent-symbolic',
    title: _("Clear the list of last used documents?"),
    subtitle: null,
    body: _("If you clear the list of last used documents, you clear the following:\n\u2022 All items from the Places \u2192 Last used documents menu item\n\u2022 All items from the last used documents list in all applications"),
    actions: [
        { signal: 'empty', label: _("Clear list") },
    ],
};

const deleteTrashEntryDialogContent = {
    icon: 'user-trash-symbolic',
    title: _("If you delete an object, it will be permanently lost."),
    subtitle: value => {
        return _("Object: %s").format(value);
    },
    body: _("Are you sure you want to permanently delete this object from the trash?"),
    actions: [
        { signal: 'remove', label: _("Delete Permanently") },
    ],
};

const ejectVolumeDialogContent = {
    icon: 'media-eject-symbolic',
    title: _("Eject the selected volume?"),
    subtitle: value => {
        return _("Volume: %s").format(value);
    },
    body: null,
    actions: [
        { signal: 'eject', label: _("Eject") },
    ],
};

const emptyTrashDialogContent = {
    icon: 'user-trash-symbolic',
    title: _("Empty all items from Trash?"),
    subtitle: null,
    body: _("All items in the Trash will be permanently deleted."),
    actions: [
        { signal: 'empty', label: _("Empty the Trash") },
    ],
};

const hibernateDialogContent = {
    icon: 'document-save-symbolic',
    title: _("Hibernate"),
    subtitle: null,
    body: _("Do you really want to hibernate the system?"),
    actions: [
        { signal: 'hibernate', label: _("Hibernate") },
    ],
};

const hibernationDialogContent = {
    icon: 'document-save-symbolic',
    title: _("Hibernate or Hybrid-Sleep"),
    subtitle: null,
    body: _("Do you really want to hibernate or hybrid sleep the system?"),
    actions: [
        { signal: 'hybrid-sleep', label: _("Hybrid-Sleep") },
        { signal: 'hibernate', label: _("Hibernate") },
    ],
};

const hybridSleepDialogContent = {
    icon: 'document-save-as-symbolic',
    title: _("Hybrid-Sleep"),
    subtitle: null,
    body: _("Do you really want to hybrid sleep the system?"),
    actions: [
        { signal: 'hybrid-sleep', label: _("Hybrid-Sleep") },
    ],
};

const removeRecentDocumentDialogContent = {
    icon: 'document-open-recent-symbolic',
    title: _("Remove the selected document?"),
    subtitle: value => {
        return _("Document: %s").format(value);
    },
    body: _("Are you sure you want to remove this document from the list of last used documents?"),
    actions: [
        { signal: 'remove', label: _("Remove") },
    ],
};

const DialogContent = {
    'clearRecentDocuments': clearRecentDocumentsDialogContent,
    'deleteTrashEntry': deleteTrashEntryDialogContent,
    'ejectVolume': ejectVolumeDialogContent,
    'emptyTrash': emptyTrashDialogContent,
    'hibernate': hibernateDialogContent,
    'hibernation': hibernationDialogContent,
    'hybridSleep': hybridSleepDialogContent,
    'removeRecentDocument': removeRecentDocumentDialogContent,
};

var ModalUserDialog = new Lang.Class({
    Name: 'ModalUserDialog',
    Extends: ModalDialog.ModalDialog,

    _init(dialog, value) {
        this.parent({ styleClass: 'modal-user-dialog' });
        let { icon, title, subtitle, body, actions } = DialogContent[dialog];
        let content = new Dialog.MessageDialogContent({
            icon: new Gio.ThemedIcon({ name: icon }),
            title: title,
            subtitle: value ? subtitle(value) : subtitle,
            body: body,
        });
        this.contentLayout.add_actor(content);
        let buttons = [{
            label: _("Cancel"),
            action: () => { this.close(); },
            key: Clutter.Escape,
        }];
        actions.forEach(action => {
            buttons.push({
                label: action.label,
                action: () => {
                    this.close();
                    let signalId = this.connect('closed', () => {
                        this.disconnect(signalId);
                        this.emit(action.signal);
                    });
                },
            });
        });
        this.setButtons(buttons);
    }
});
